#include <windows.h>
#include <stdio.h>

char pcsFormattedNumber[64];  // 20 digits with a comma every 3 places uses about 30 characters; double that for safety

// Returns a string representation (including thousands separators) of the specified DWORDLONG value
char* FormatNumber(DWORDLONG dwlNumber)
{
	// GetNumberFormat needs a string input
	char pcsInput[21];  // ((DWORDLONG)-1) has 20 digits
	if(-1 != sprintf_s(pcsInput, sizeof(pcsInput)/sizeof(pcsInput[0]), "%I64u", dwlNumber))
	{
		// Format the input according to the user's settings
		if(0 < GetNumberFormat(LOCALE_USER_DEFAULT, 0, pcsInput, NULL, pcsFormattedNumber, sizeof(pcsFormattedNumber)/sizeof(pcsFormattedNumber[0])))
		{
			// GetNumberFormat doesn't provide a way to easily avoid unnecessary decimal points/digits...
			char pcsDecimal[4];  // Documented maximum length of LOCALE_SDECIMAL
			if(0 < GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SDECIMAL, pcsDecimal, sizeof(pcsDecimal)/sizeof(pcsDecimal[0])))
			{
				// ... so remove them if present
				char* const pcsEnd = strstr(pcsFormattedNumber, pcsDecimal);
				if(NULL != pcsEnd)
				{
					*pcsEnd = '\0';
				}
			}
			// Return the formatted string
			return pcsFormattedNumber;
		}
	}
	// Error, return a placeholder
	return "?";
}

int main(int argc, char* argv[])
{
	// Validate arguments
	if((3 == argc) && (2 == strlen(argv[1])) && isalpha(argv[1][0]) && (':' == argv[1][1]))
	{
		// Validate drive type
		const char cDrive = (char)toupper(argv[1][0]);
		char pcsDriveRoot[] = "?:\\";
		pcsDriveRoot[0] = cDrive;
		if(DRIVE_CDROM == GetDriveType(pcsDriveRoot))
		{
			// Output banner
			const char* const pcsOutputFile = argv[2];
			printf("\nExtracting ISO from %c: to \"%s\"\n", cDrive, pcsOutputFile);
			char pcsInputFile[] = "\\\\.\\?:";
			pcsInputFile[4] = cDrive;

			// Open handle for DVD drive volume
			BOOL bSuccess = FALSE;
			// Use FILE_FLAG_NO_BUFFERING per CreateFile recommendations for opening volumes
			const HANDLE hInput = CreateFile(pcsInputFile, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL);
			if(INVALID_HANDLE_VALUE != hInput)
			{
				// Open handle for output file
				const HANDLE hOutput = CreateFile(pcsOutputFile, GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL);
				if(INVALID_HANDLE_VALUE != hOutput)
				{
					// Try to resume extraction if possible and relevant
					DWORDLONG dwlTotalBytes = 0;
					LARGE_INTEGER liZero = { 0 };
					LARGE_INTEGER liNewFilePointer = { 0 };
					if(SetFilePointerEx(hOutput, liZero, &liNewFilePointer, FILE_END))
					{
						if(0 < liNewFilePointer.QuadPart)
						{
							// Some data already extracted, try to resume
							if(SetFilePointerEx(hInput, liNewFilePointer, 0, FILE_BEGIN))
							{
								dwlTotalBytes += liNewFilePointer.QuadPart;
								printf("Resuming extraction from byte position %s\n", FormatNumber(dwlTotalBytes));
							}
							else
							{
								// Unable to resume; start over
								(void)SetFilePointerEx(hOutput, liZero, 0, FILE_BEGIN);
								(void)SetEndOfFile(hOutput);
							}
						}
					}
					// Allocate data buffer
					const DWORD dwReadWriteSize = 2048*512;  // 2048 is CD sector size; 512 is large buffer
					const LPVOID pbBuffer = VirtualAlloc(NULL, dwReadWriteSize, MEM_COMMIT, PAGE_READWRITE);
					if(NULL != pbBuffer)
					{
						// Extract all data
						bSuccess = true;  // Assume success
						DWORD dwBytesRead = 1;  // Any non-0 value will do
						const int MaxFailedReadRetries = 10;
						int iReadRetriesRemaining = MaxFailedReadRetries;
						while(bSuccess && (0 != dwBytesRead))  // While successful and not EOF...
						{
							// Read data
							if((bSuccess = ReadFile(hInput, pbBuffer, dwReadWriteSize, &dwBytesRead, NULL)) && (0 < dwBytesRead))
							{
								// Write data
								iReadRetriesRemaining = MaxFailedReadRetries;
								DWORD dwBytesWritten;
								if((bSuccess = WriteFile(hOutput, pbBuffer, dwBytesRead, &dwBytesWritten, NULL)) && (dwBytesWritten == dwBytesRead))
								{
									// Update status
									dwlTotalBytes += dwBytesRead;
									printf("%s bytes extracted...\r", FormatNumber(dwlTotalBytes));
								}
							}
							else
							{
								// Retry the read
								if(!bSuccess && (ERROR_CRC == GetLastError()) && (0 < iReadRetriesRemaining))
								{
									printf("\nRead failed due to CRC check failure, retrying...\n");
									iReadRetriesRemaining--;
									bSuccess = true;  // Fake success to retry
									dwBytesRead = 1;  // Any non-0 value will do
								}
							}
						}
						(void)VirtualFree(pbBuffer, 0, MEM_RELEASE);
					}
					(void)CloseHandle(hOutput);
					if(!bSuccess)
					{
						// NOT deleting incomplete output file to allow for attempted resume
						// (void)DeleteFile(pcsOutputFile);
					}
				}
				(void)CloseHandle(hInput);
			}

			// Output results
			if(bSuccess)
			{
				printf("\nDone.\n");
			}
			else
			{
				const DWORD dwLastError = GetLastError();
				LPTSTR ptsMessage;
				if(0 < FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER, 0, dwLastError, 0, (LPTSTR)(&ptsMessage), 0, 0))
				{
					printf("\nError %d: %s\n", dwLastError, ptsMessage);
					(void)LocalFree(ptsMessage);
				}
				printf("NOTE: Partial file has NOT been deleted so that recovery can be attempted.\n");
			}
		}
		else
		{
			printf("\nError: %c: is not a CD/DVD drive\n", cDrive);
		}
	}
	else
	{
		printf(
		//  "12345678901234567890123456789012345678901234567890123456789012345678901234567890"
			"\n"
			"ExtractISO - Creates an ISO image file from a CD/DVD\n"
			"\n"
			"ExtractISO reads raw sectors from the specified CD/DVD volume and writes them\n"
			"to the specified ISO image file.  The resulting file can be burned to a blank\n"
			"CD/DVD disc, mounted in a virtual CD/DVD drive, or opened in an ISO viewer.\n"
			"\n"
			"For a discussion of the limitations of this process, please refer to:\n"
			"   http://www.cdrfaq.org/faq02.html#S2-28\n"
			"\n"
			"Note that direct access to a CD/DVD volume requires administrative privileges.\n"
			"\n"
			"ExtractISO attempts to read damaged media by retrying failed reads a few times\n"
			"before giving up. In the event of a persistent read failure, the partial output\n"
			"file is not deleted so that subsequent extraction attempts (possibly after\n"
			"cleaning the media and/or using a different drive to read it) can be attempted\n"
			"without the loss of any successfully extracted data.\n"
			"\n"
			"Syntax: ExtractISO D: File.iso\n"
			"   D: is the CD/DVD drive to extract from\n"
			"   File.iso is the file to write to\n"
		);
	}
	return 0;
}
